;; Generated package description from gnosis.el  -*- no-byte-compile: t -*-
(define-package "gnosis" "0.5.3" "Spaced Repetition System" '((emacs "27.2") (emacsql "4.1.0") (compat "29.1.4.2") (transient "0.7.2") (org-gnosis "0.0.9")) :commit "4b87d58574d1b8cb2eb7bf7fe571efa758838c03" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("extensions") :url "https://thanosapollo.org/projects/gnosis")
